@@ -1,22 +1,21 @@
 """Classes and functions related to the Logger part of LDL."""
 
+import logging
 from time import sleep
 
+import influxdb
 import rpyc
-from influxdb import InfluxDBClient
 
-# pylint: disable=no-name-in-module
-from multiprocess import (
-    Process,
-    Queue,
-    Value,
-    Event,
-)
+from multiprocessing import Event, Process, Queue, Value
+
+from .utils import parse_netloc
+
+debug_logger = logging.getLogger("lab_data_logger.logger")
 
 rpyc.core.protocol.DEFAULT_CONFIG["allow_pickle"] = True
 
 JOIN_TIMEOUT = 1  # timeout for joining processes
-LOGGER_SHOW_INTERVAL = 0.5
+LOGGER_SHOW_INTERVAL = 0.5  # update intervall for show_logger_status
 
 
 class Puller:
@@ -32,44 +31,27 @@ class Puller:
     port : int
         Port at which the DataService can be accessed (default 18861).
     measurement : str
-        Name of the measurement. This name will be used as the measurement when writing
-        to an InfluxDB (default "test").
+        Name of the measurement. This name will be used as the measurement when
+        writing to an InfluxDB.
     interval : float
         Logging interval in seconds.
     """
 
-    def __init__(self, queue, host, port, measurement, interval):
+    def __init__(self, queue, host, port, measurement, interval, fields=None):
         self.queue = queue
         self.host = host
         self.port = port
         self.measurement = measurement
         self.interval = interval
+        self.fields = fields
         # shared value for communicating the processes status
         self._shared_counter = Value("i", -1)
         self.stop_event = Event()
         self.pull_process = Process(
-            target=self._pull, args=(self.queue, self._shared_counter, self.stop_event)
+            target=self._pull,
+            args=(self.queue, self._shared_counter, self.stop_event, self.fields),
         )
 
-    def start_process(self):
-        """Start process continously pulling data and writing it to the queue."""
-        self.pull_process.start()
-
-    def stop_process(self):
-        """
-        Stop process continously pulling data and writing it to the queue.
-
-        Returns
-        -------
-        exitcode : int
-        """
-        self.stop_event.set()
-        self.pull_process.join(JOIN_TIMEOUT)
-        if self.pull_process.is_alive():
-            self.pull_process.terminate()
-        exitcode = self.pull_process.exitcode
-        return exitcode
-
     @property
     def counter(self):
         """
@@ -77,27 +59,38 @@ class Puller:
         """  # noqa D401
         return self._shared_counter.value
 
-    def _pull(self, queue, shared_counter, stop_event):
+    def _pull(self, queue, shared_counter, stop_event, fields=None):
         # the worker of the pulling process
         try:
             service = rpyc.connect(self.host, self.port)
-            print(
-                "Connected to {} on port {}".format(
-                    service.root.get_service_name(), self.port
-                )
+            debug_logger.info(
+                f"Connected to {service.root.get_service_name()} on port {self.port}."
             )
-        except ConnectionRefusedError as error:
-            raise ConnectionRefusedError(
-                "Connection to service at {}:{} refused".format(self.host, self.port)
-            ) from error
+
+        except ConnectionRefusedError:
+            debug_logger.exception(
+                f"Connection to service at {self.host}:{self.port} refused."
+            )
+            debug_logger.warning(f"Stopping pull process from {self.host}:{self.port}.")
+            stop_event.set()
         else:
             shared_counter.value += 1  # change from -1 to 0
+            # worker loop
             while not stop_event.is_set():
-                data = service.root.exposed_get_data()
-                data["measurement"] = self.measurement
-                queue.put([data])
-                shared_counter.value += 1
-                sleep(self.interval)
+                try:
+                    data = service.root.exposed_get_data(fields=fields)
+                    data["measurement"] = self.measurement
+                    queue.put([data])
+                    shared_counter.value += 1
+                    sleep(self.interval)
+                except EOFError:
+                    debug_logger.error(
+                        f"Connection to {self.host}:{self.port} closed by peer."
+                    )
+                    debug_logger.warning(
+                        f"Stopping pull process from {self.host}:{self.port}."
+                    )
+                    stop_event.set()
 
 
 class Pusher:
@@ -125,18 +118,24 @@ class Pusher:
         self.host = host
         self.port = port
         self.database = database
-        self.influxdb_client = InfluxDBClient(host, port, user, password, database)
-
-        # shared value for communicating the processes status
-        self._shared_counter = Value("i", -1)
-
-        self.push_process = Process(
-            target=self._push, args=(self.queue, self._shared_counter)
+        self.influxdb_client = influxdb.InfluxDBClient(
+            host, port, user, password, database
         )
 
-    def start_process(self):
-        """Start process continously reading the queue and writing to the InfluxDB."""
-        self.push_process.start()
+        # check connection
+        available_databases = self.influxdb_client.get_list_database()
+        available_databases = [item["name"] for item in available_databases]
+        if self.database in available_databases:
+            # shared value for communicating the processes status
+            self._shared_counter = Value("i", -1)
+
+            self.push_process = Process(
+                target=self._push, args=(self.queue, self._shared_counter)
+            )
+        else:
+            raise influxdb.exceptions.InfluxDBClientError(
+                f"No database named '{self.database}' found. Make sure it exists."
+            )
 
     @property
     def counter(self):
@@ -149,13 +148,19 @@ class Pusher:
         shared_counter.value += 1  # change from -1 to 0
         while True:
             data = queue.get()
-            self.influxdb_client.write_points(data)
+            try:
+                self.influxdb_client.write_points(data)
+            except influxdb.exceptions.InfluxDBClientError:
+                # FIXME: Change behaviour depending on which error is thrown.
+                debug_logger.exception(f"Could not write data {data} to the database.")
+
             shared_counter.value += 1
 
 
 class Logger(rpyc.Service):
     """
-    Service comprised of a Pusher and a number of Pullers and methods for managing them.
+    Service comprised of a Pusher and a number of Pullers and methods for
+    managing them.
 
     Parameters
     ----------
@@ -171,33 +176,17 @@ class Logger(rpyc.Service):
         Name of the database that should be used.
     """
 
-    def __init__(self, host, port, user, password, database):
+    def __init__(
+        self, host="localhost", port=8086, user=None, password=None, database=None
+    ):
         super(Logger, self).__init__()
         self.queue = Queue()
-        self.start_pusher_process(host, port, user, password, database)
+        self.pusher = Pusher(self.queue, host, port, user, password, database)
+        self.pusher.push_process.start()
+        debug_logger.debug("Pusher process started.")
         self.exposed_pullers = {}
 
-    def start_pusher_process(self, host, port, user, password, database):
-        """
-        Start the Pusher process.
-
-        Parameters
-        ----------
-        host : str
-            Hostname of the InfluxDB.
-        port : int
-            Port of the InfluxDB.
-        user : str
-            Username of the InfluxDB.
-        password : str
-            Password for the InfluxDB.
-        database : str
-            Name of the database that should be used.
-        """
-        self.pusher = Pusher(self.queue, host, port, user, password, database)
-        self.pusher.start_process()
-
-    def exposed_add_puller(self, host, port, measurement, interval):
+    def exposed_add_puller(self, host, port, measurement, interval, fields=None):
         """
         Add and start a Puller process.
 
@@ -208,40 +197,53 @@ class Logger(rpyc.Service):
         port : int
             Port at which the DataService can be accessed (default 18861).
         measurement : str
-            Name of the measurement. This name will be used as the measurement when writing
-            to an InfluxDB (default "test").
+            Name of the measurement. This name will be used as the measurement
+            when writing to an InfluxDB.
         interval : float
             Logging interval in seconds.
         """
         netloc = f"{host}:{port}"
         if netloc in self.exposed_pullers.keys():
-            raise Exception(f"{netloc} is already being pulled.")
-        try:
-            puller = Puller(self.queue, host, port, measurement, interval)
-            puller.start_process()
-        except ConnectionRefusedError as error:
-            # FIXME: this doesn't do anything yet, child processes don't communicate errors to the
-            # parent
-            raise error
+            debug_logger.error(f"{netloc} is already being pulled.")
         else:
+            puller = Puller(
+                self.queue, host, port, measurement, interval, fields=fields
+            )
+            debug_logger.info(f"Starting pull process for {netloc}.")
+            puller.pull_process.start()
             self.exposed_pullers[netloc] = puller
 
     def exposed_remove_puller(self, netloc):
         """
         Stop and remove a puller from the logger.
 
-        netloc : str
-            Network location, e.g. localhost:18861
+        netloc : str or int
+            Network location, e.g. localhost:18861. If an int is passed,
+            localhost is assumed.
         """
+        host, port = parse_netloc(netloc)
+        netloc = f"{host}:{port}"
         try:
-            exitcode = self.exposed_pullers[netloc].stop_process()
-            print(f"Puller process for {netloc} exited with code {exitcode}.")
+            puller = self.exposed_pullers[netloc]
+            # shutting down via Event
+            puller.stop_event.set()
+            puller.pull_process.join(JOIN_TIMEOUT)
+            if puller.pull_process.is_alive():
+                # if not successful, terminate
+                puller.pull_process.terminate()
+            debug_logger.info(
+                "Puller process for {} exited with code {}.".format(
+                    netloc, puller.pull_process.exitcode
+                )
+            )
             del self.exposed_pullers[netloc]
-        except KeyError as error:
-            raise KeyError(f"No Puller pulling from {netloc}") from error
+        except KeyError:
+            debug_logger.error(f"No Puller pulling from {netloc}")
 
     def exposed_get_display_text(self):
-        """Print status of connected DataServices and the InfluxDB, continously."""
+        """
+        Print status of connected DataServices and the InfluxDB, continously.
+        """
         display_text = "\nLAB DATA LOGGER\n"
         display_text += "Logging to {} on {}:{} (processed entry {}).\n".format(
             self.pusher.database,
@@ -289,7 +291,7 @@ def start_logger(logger_port, host, port, user, password, database):
 
     proc = Process(target=threaded_server.start)
     proc.start()
-    print("Started logger on port {}.".format(logger_port))
+    debug_logger.info(f"Started logger on port {logger_port}.")
 
 
 def _get_logger(port):
@@ -311,12 +313,12 @@ def _get_logger(port):
     except ConnectionRefusedError as error:
         raise ConnectionRefusedError(
             "Connection to Logger refused."
-            "Make sure there a Logger is running on port {}.".format(port),
+            f"Make sure there a Logger is running on port {port}.",
         ) from error
     return logger
 
 
-def add_puller_to_logger(logger_port, netloc, measurement, interval):
+def add_puller_to_logger(logger_port, netloc, measurement, interval, fields=None):
     """
     Add a Puller to a Logger.
 
@@ -332,10 +334,13 @@ def add_puller_to_logger(logger_port, netloc, measurement, interval):
         Name of the measurement, i.e. the measurement field of InfluxDB.
     interval : int
         Logging interval in seconds.
+    fields : list
+        Optional list of fields that should be returned. If more fields are
+        provided by the service they will be filtered.
     """
     logger = _get_logger(logger_port)
-    host, port = _parse_netloc(netloc)
-    logger.root.exposed_add_puller(host, port, measurement, interval)
+    host, port = parse_netloc(netloc)
+    logger.root.exposed_add_puller(host, port, measurement, interval, fields=fields)
 
 
 def remove_puller_from_logger(logger_port, netloc):
@@ -368,19 +373,3 @@ def show_logger_status(logger_port):
         display_text = logger.root.exposed_get_display_text()
         print(display_text)
         sleep(LOGGER_SHOW_INTERVAL)
-
-
-def _parse_netloc(netloc):
-    # Split network location pair hostname:port into
-    split_netloc = netloc.split(":")
-    if len(split_netloc) == 2:
-        # host:port pair
-        host = split_netloc[0]
-        port = int(split_netloc[1])
-    elif len(split_netloc) == 1:
-        # only port
-        host = "localhost"
-        port = int(split_netloc[0])
-    else:
-        raise ValueError("'{}' is not a valid location".format(netloc))
-    return host, port
